{# Answering Instructions #}
{% if use_code %}
Please respond only with a comma-separated list of the code of the raked options, with square brackets. E.g., [0, 1, 3]
{% else %}
Please respond only with a comma-separated list of the ranked options, with square brackets. E.g., ['Good', 'Bad', 'Ugly']
{% endif %}
{% if include_comment %}
After the answer, you can put a comment explaining your choice on the next line.
{% endif %}


