    {{question_text}}
        {% for option in question_options %}
            {{option}} : {{ option_labels.get(option, "") }}
        {% endfor %}
    Return a valid JSON formatted like this, selecting only the code of the option (codes start at 0):
    {% if include_comment %}  
        {"answer": <put answer code here>, "comment": <comment>} 
    {% else %} 
        {"answer": <put answer here>} 
    {% endif %}
    Only 1 option may be selected.
