"""Base module wrapper for backward compatibility.

This module re-exports everything from the edsl.base module.
"""

# Re-export everything from edsl.base module
from edsl.base import *
from edsl.base.base_class import *