{# Question Presention #}
{{question_text}}

{% if use_code %}
    {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
    {% endfor %}
{% else %}
    {% for option in question_options %}
        {{option}}
    {% endfor %}
{% endif %}

Only 1 option may be selected.

{# Answering Instructions #}
Return a valid JSON formatted like this:

{% if use_code %}
    {% if include_comment %}
        {"answer": <put answer code here>, "comment": "<put explanation here>"}
    {% else %}
        {"answer": <put answer code here>}
    {% endif %}
{% else %}

    {% if include_comment %}
        {"answer": <text of option>, "comment": "<put explanation here>"}
    {% else %}
        {"answer": <put option here>}
    {% endif %}

{% endif %}
