<div class="overview-section">
    <h2>Exceptions Report</h2>
    <p>
        This report summarizes exceptions encountered in the job that was run.
    </p>

    <table class="summary-table">
        <tbody>
            <tr>
                <td><strong>Total interviews</strong></td>
                <td>{{ interviews|length }}</td>
            </tr>
            <tr>
                <td><strong>Interviews with exceptions</strong></td>
                <td>{{ num_exceptions }}</td>
            </tr>
        </tbody>
    </table>
    <p>
        An interview is the result of one survey, taken by one agent, with one model and one scenario (if any).
    </p>
</div>